/* Shared library add-on to iptables to add inputdev match support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_inputdev.h>

#undef debug
/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"inputdev v%s options:\n"
" --devs name[,name,name...]\n"
" --devs ! name[,name,name...]\n",
IPTABLES_VERSION);
}

//match command like -m inputdev --devs eth0,eth1,eth2 -j DROP
static struct option opts[] = {
	{ "devs", 1, 0, '1' },
	{0}
};

static unsigned int
parse_multi_devs(const char *devstring, char (*devs)[IFNAMSIZ])
{
	char *buffer, *cp, *next;
	unsigned int i;
#ifdef debug
printf("start to analyse devs %s \n", devstring);
#endif
	buffer = strdup(devstring);
	if (!buffer) exit_error(OTHER_PROBLEM, "strdup failed");

	for (cp = buffer, i = 0; cp && i < IPT_MULTI_INPUTDEVS; cp = next, i++)
	{
		next = strchr(cp, ',');
		if(next) *next++ = '\0';
		
		if(strlen(cp) >= IFNAMSIZ)
		{
			free(buffer);
			exit_error(PARAMETER_PROBLEM, 
				"invalid dev name specified, len of name should little than 16");
		}
		strcpy(devs[i], cp);
#ifdef debug
printf("set devs[%d] to %s: %s\n", i ,cp, devs[i]);
#endif
	}
	if(cp)
	{
		free(buffer);
		exit_error(PARAMETER_PROBLEM, "too many devs specified");
	}
	free(buffer);
	return i;
}


/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, 
			char **argv, 
			int invert,
			unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_inputdev *multiinfo
		= (struct ipt_inputdev *)(*match)->data;
		
	multiinfo->invert = 0;
	switch (c) {
	case '1':
		check_inverse(optarg, &invert, &optind, 0);
		multiinfo->count = parse_multi_devs(argv[optind - 1], multiinfo->devs);
		
		if (invert)	multiinfo->invert = 1;						     
		break;

	default:
		return 0;
	}

	if (*flags)
		exit_error(PARAMETER_PROBLEM,
			   "inputdev can only have one option");
	*flags = 1;
	return 1;
}

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
}

/* Final check; must specify something. */
static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM, "inputdev expection an option");
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	const struct ipt_inputdev *multiinfo
		= (const struct ipt_inputdev *)match->data;
	unsigned int i;
	
	if(multiinfo->invert)printf("inputdev ! ");
	else printf("inputdev ");
		
	for (i = 0; i < multiinfo->count; i++) {
		printf("%s", i ? "," : "");
		printf("%s", multiinfo->devs[i]);
	}
	printf(" ");
}


/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	const struct ipt_inputdev *multiinfo
		= (const struct ipt_inputdev *)match->data;
	unsigned int i;

	if(multiinfo->invert)printf("--devs ! ");
	else printf("--devs ");

	for (i=0; i < multiinfo->count; i++) {
		printf("%s", i ? "," : "");
		printf("%s", multiinfo->devs[i]);
	}
	printf(" ");
}

static
struct iptables_match inputdev
= { NULL,
    "inputdev",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_inputdev)),
    IPT_ALIGN(sizeof(struct ipt_inputdev)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void
_init(void)
{
	register_match(&inputdev);
}
